@@ -80,11 +80,13 @@ module System.Process.Typed
       -- * Exceptions
     , ExitCodeException (..)
     , ByteStringOutputException (..)
+      -- * Unsafe functions
+    , unsafeProcessHandle
     ) where
 
 import qualified Data.ByteString as S
 import Data.ByteString.Lazy.Internal (defaultChunkSize)
-import Control.Exception (throwIO)
+import Control.Exception (assert, evaluate, throwIO)
 import Control.Monad (void)
 import Control.Monad.IO.Class
 import qualified System.Process as P
@@ -620,7 +622,9 @@ startProcess pConfig'@ProcessConfig {..} = liftIO $ do
                 -- then call waitForProcess ourselves
                 Left _ -> do
                     P.terminateProcess pHandle
-                    void $ P.waitForProcess pHandle
+                    ec <- P.waitForProcess pHandle
+                    success <- atomically $ tryPutTMVar pExitCode ec
+                    evaluate $ assert success ()
 
     return Process {..}
   where
@@ -822,3 +826,25 @@ instance Show ExitCodeException where
 data ByteStringOutputException = ByteStringOutputException SomeException (ProcessConfig () () ())
     deriving (Show, Typeable)
 instance Exception ByteStringOutputException
+
+-- | Take 'System.Process.ProcessHandle' out of the 'Process'.
+-- This method is needed in cases one need to use low level functions
+-- from the @process@ package. Use cases for this method are:
+--
+--   1. Send a special signal to the process.
+--   2. Terminate the process group instead of terminating single process.
+--   3. Use platform specific API on the underlying process.
+--
+-- This method is considered unsafe because the actions it performs on
+-- the underlying process may overlap with the functionality that
+-- @typed-process@ provides. For example the user should not call
+-- 'System.Process.waitForProcess' on the process handle as eiter
+-- 'System.Process.waitForProcess' or 'stopProcess' will lock.
+-- Additionally, even if process was terminated by the
+-- 'System.Process.terminateProcess' or by sending signal,
+-- 'stopProcess' should be called either way in order to cleanup resources
+-- allocated by the @typed-process@.
+--
+-- @since 0.1.1
+unsafeProcessHandle :: Process stdin stdout stderr -> P.ProcessHandle
+unsafeProcessHandle = pHandle
